"""This module contain functions that used to visualized the analyzed data."""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from typing import List, Dict, Any, Union
from . import auxiliary as aux
from .dataset import LoadData


def heatmap(
    dataset: LoadData,
    min_range: float,
    max_range: float,
    export_data: str = None,
    display_rxn_time: bool = False,
    export_fig: str = None,
    plot_distance: bool = False,
    lower_limit: float = None,  
    upper_limit: float = None,  
) -> None:
    """
    Plots a heatmap based on the intensity of peaks as a function of the q range and scan number.

    :param dataset: Dataset object of the associated experiment.
    :param min_range: Minimum range of q values.
    :param max_range: Maximum range of q values.
    :param export_data: Whether or not to export data in Excel format.
    :param display_rxn_time: If True, display reaction time. If False, display scan number.
    :param export_fig: If the path is given, export graph to the specified path.
    :param plot_distance: If True, plot the heatmap in real distance rather than in arbitrary position.
    :param lower_limit: Minimum intensity value to display in the heatmap.
    :param upper_limit: Maximum intensity value to display in the heatmap.
    """
    max_positions = 0
    height_group_frame = dataset.height_group_frame
    fl_num = dataset.fl_num
    height_group = dataset.height_group
    integrated_data = dataset.fl_integrated
    for x_val in height_group_frame:
        Y = aux.get_data(
            fl=integrated_data,
            dataset_path=f"{x_val}.1/p3_integrate/integrated/intensity",
        )
        max_positions = max(max_positions, len(Y))

    x = np.linspace(0, max_positions - 1, max_positions)
    if display_rxn_time:
        array_timestamp = [
            aux.get_scan_time(fl=dataset.fl_raw, scan_num=f) for f in height_group_frame
        ]
        array_rxn_time = [(t - array_timestamp[0]) / 60 for t in array_timestamp]
        y = np.array(array_rxn_time)
    else:
        y = np.array(height_group_frame)
    x, y = np.meshgrid(x, y)
    z = np.empty_like(x)

    for i, y_val in enumerate(height_group_frame):
        X = aux.get_data(
            fl=integrated_data, dataset_path=f"{y_val}.1/p3_integrate/integrated/q"
        )
        Y = aux.get_data(
            fl=integrated_data,
            dataset_path=f"{y_val}.1/p3_integrate/integrated/intensity",
        )

        for j, x_val in enumerate(x[i]):
            if 0 <= x_val < len(Y):
                z[i, j] = aux.find_peak_height(
                    X=X, Y=Y[int(x_val)], x_min=min_range, x_max=max_range
                )
            else:
                z[i, j] = np.nan
    if plot_distance:
        distance_multiplier = aux.get_height_diff(dataset)
    else:
        distance_multiplier = 1
    plt.imshow(
        z.T,
        extent=[y.min(), y.max(), 0, max_positions * distance_multiplier],
        origin="lower",
        aspect="auto",
        cmap="RdYlBu",
        interpolation="nearest",
        vmin=lower_limit, 
        vmax=upper_limit,  
    )

    plt.colorbar(label="Maximum peak height")
    if display_rxn_time:
        x_label = "Time (min)"
    else:
        x_label = "Scan number"
    plt.xlabel(x_label)
    if plot_distance:
        y_label = "Distance (μm)"
    else:
        y_label = "Position"
    plt.ylabel(y_label)
    plt.title(
        f"Exp: {fl_num}, height group: {height_group}, q range = [{min_range:.4f},{max_range:.4f}]",
        size=11,
    )
    ax = plt.gca()
    ax.minorticks_on()
    y_ticks = np.linspace(0, max_positions * distance_multiplier, 11)
    ax.set_yticks(y_ticks)
    ax.yaxis.set_major_locator(ticker.FixedLocator(y_ticks))
    if export_fig:
        plt.savefig(export_fig)
    plt.show()
    if export_data:
        data = {
            x_label: y.flatten(),
            y_label: x.flatten() * distance_multiplier,
            "Maximum peak height": z.flatten(),
        }
        df = pd.DataFrame(data)
        df.to_excel(export_data, index=False)



def compare_peak_fe(
    dataset: LoadData,
    x_min: float,
    x_max: float,
    position_range: Union[int, List[int]],
    path_gc_excel: str,
    smoothing_window: Union[None, int] = None,
    n_pol: int = 3,
    compare_product: str = "C2H4",
    compare_product_label: str = "C$_2$H$_4$",
    axis_y_log: bool = False,
    axis_y_min: Union[bool, int] = False,
    axis_y_max: Union[bool, int] = False,
    axis_x_min: Union[bool, int] = False,
    axis_x_max: Union[bool, int] = False,
    export_table: Union[bool, str] = False,
) -> None:
    """
    Function to plot the X-ray intensity and the Faradaic efficiency for H2 and C2H4 (for Cu) or CO (For Ag).

    This function also includes a built-in smoothing function for the X-ray data and the ability to export the X-ray
    data and the FE into an excel file.
    """
    fl_num = int(dataset.fl_num)
    height_group = dataset.height_group
    # Convert single position to a list
    if isinstance(position_range, int):
        position_range = [position_range]

    # Collect data for all positions
    dfs_xray = []
    for pos in range(position_range[0], position_range[-1] + 1):
        dfs_xray.append(
            aux.get_peak_height_time(
                dataset, x_min, x_max, pos, smoothing_window, n_pol
            )
        )

    # Average the data
    avg_df_xray = dfs_xray[0].copy()
    for df in dfs_xray[1:]:
        avg_df_xray["peak height"] += df["peak height"]
    avg_df_xray["peak height"] /= len(dfs_xray)

    df_fe = aux.get_fe(path_gc_excel)
    x_0 = df_fe["time"][0]

    fig, ax1 = plt.subplots()

    # Plot the X-ray intensity
    y1_axis_color = "royalblue"
    ax1.minorticks_on()
    ax1.tick_params(axis="y", which="both", colors=y1_axis_color)
    ax1.set_xlabel("Time (min)")
    ax1.set_ylabel("Intensity", color=y1_axis_color)
    ax1.tick_params(axis="y", colors=y1_axis_color)
    ax1.spines["left"].set_color(y1_axis_color)
    ax1.plot(
        (avg_df_xray["time"] - x_0) / 60,
        avg_df_xray["peak height"],
        color=y1_axis_color,
        label="Average X-ray peak height",
    )
    if axis_y_min is not False and axis_y_max is not False:
        ax1.set_ylim(axis_y_min, axis_y_max)
    if axis_x_min is not False and axis_x_max is not False:
        ax1.set_xlim(axis_x_min, axis_x_max)
    ax1.legend(loc=2)
    if axis_y_log:
        ax1.set_yscale("log")

    # Plot the FE
    time_adjusted = (df_fe["time"] - x_0) / 60
    ax2 = ax1.twinx()
    ax2.minorticks_on()
    ax2.plot(time_adjusted, df_fe["H2"] * 100, color="orangered", label="H$_2$")
    ax2.plot(
        time_adjusted,
        df_fe[compare_product] * 100,
        color="forestgreen",
        label=compare_product_label,
    )
    ax2.set_ylabel("Faradaic efficiency (%)")
    ax2.spines["left"].set_color(y1_axis_color)
    ax2.legend(loc=1)
    plt.title(
        f"Exp: {fl_num}, height group: {height_group},peak range=[{x_min},{x_max}], pos. = {position_range}"
    )

    # Export the table for further plotting
    if export_table is not False:
        len_xray = len(avg_df_xray)
        len_fe = len(df_fe)

        max_len = max(len_xray, len_fe)

        if len_xray < max_len:
            avg_df_xray = avg_df_xray.reindex(range(max_len))
        if len_fe < max_len:
            df_fe = df_fe.reindex(range(max_len))

        df_export = pd.DataFrame(
            {
                "X-ray time/min": (avg_df_xray["time"] - x_0) / 60,
                "Average X-ray peak intensity": avg_df_xray["peak height"],
                "FE_time/min": time_adjusted,
                "FE_H2 / %": df_fe["H2"] * 100,
            }
        )
        df_export[f"FE_{compare_product} / %"] = df_fe[compare_product] * 100
        df_export.to_excel(export_table, index=False)


def peak_span(
    dataset: LoadData,
    x_min: float,
    x_max: float,
    position: int,
    n_plot: int,
    export_table: Union[bool, str] = False,
) -> None:
    """
    Function to plot peaks in the designated range at specific scan interval to check the peak broadening.

    :param datasey: dataset object of the associated experiment.
    :param x_min: minimum q-range of the plotting window.
    :param x_max: maximum q-range of the plotting window.
    :param position: position of the scan.
    :param n_plot: number of plot to be overlayed. :export_table: if given, a path to export an Excel file containing
        peak information.
    """

    def select_frames(frame_list: List[int], n: int) -> List[int]:
        """
        Selects 'n' evenly spaced frames from a given list of frame numbers.

        The function calculates the step size and the starting index to ensure that
        the frames are as evenly distributed as possible across the list. If 'n' is
        greater than the number of frames, it returns the entire list.

        Parameters:
        frame_list (List[int]): The list of frame numbers to select from.
        n (int): The number of frames to select.

        Returns:
        List[int]: A list containing the selected frame numbers.

        Raises:
        ValueError: If 'n' is a non-positive integer.
        """
        if n <= 0:
            raise ValueError(
                "The number of frames to select must be a positive integer."
            )

        total_frames = len(frame_list)
        if n >= total_frames:
            return (
                frame_list  # If n is greater than the list size, return the whole list.
            )

        step = max(total_frames // n, 1)  # Ensure step is at least 1
        selected_frames = []

        # Calculate the starting index to make the distribution as even as possible
        start_index = (total_frames - (step * (n - 1))) // 2

        for i in range(n):
            # Calculate the index of the frame to be selected
            index = start_index + i * step
            # Append the selected frame to the list
            selected_frames.append(frame_list[index])

        return selected_frames

    selected_frames = select_frames(dataset.height_group_frame, n_plot)
    fig, ax = plt.subplots()
    ax.minorticks_on()
    df_export = pd.DataFrame()
    for frame in selected_frames:
        df_frame = aux.export_spectrum(data=dataset, scan_num=frame, position=position)
        idx = [
            i
            for i in range(0, len(df_frame))
            if df_frame["q"][i] > x_min and df_frame["q"][i] < x_max
        ]
        plt.plot(
            df_frame["q"][idx],
            df_frame["count"][idx],
            label=f"Scan number: {frame}",
            marker=".",
        )
        df_export.insert(
            len(df_export.columns), f"q - frame:{frame}", df_frame["q"][idx]
        )
        df_export.insert(
            len(df_export.columns), f"count - frame:{frame}", df_frame["count"][idx]
        )
    plt.legend()
    plt.xlabel("q")
    plt.ylabel("count")

    if export_table is not False:
        df_export.to_excel(export_table, index=False)


def vertical_compare(
    dataset: LoadData,
    x_min: float,
    x_max: float,
    scan_number: Union[int, List[int]],
    export_table: Union[bool, str] = False,
) -> None:
    """
    Function to plot the highest peak intensity within a given q-range as a function of position for a specified scan
    number.

    :param dataset: dataset object of the associated experiment.
    :param x_min: Minimum q-value for the selection window.
    :param x_max: Maximum q-value for the selection window.
    :param scan_number: Scan number(s) to be included in the plot (can be a single value or a list).
    :param export_table: If provided, path to export an Excel file containing the plotted data.
    """
    # Convert single scan number to a list if it's not already a list
    if isinstance(scan_number, int):
        scan_number = [scan_number]

    # Initialize a dictionary to store peak heights for each position
    peak_heights_dict = {}

    # Loop through each scan number
    for scan in scan_number:
        # Retrieve the intensity data for the given scan number
        Y = aux.get_data(
            fl=dataset.fl_integrated,
            dataset_path=f"{scan}.1/p3_integrate/integrated/intensity",
        )
        X = aux.get_data(
            fl=dataset.fl_integrated, dataset_path=f"{scan}.1/p3_integrate/integrated/q"
        )

        # Find the peak height for each position within the q-range
        for pos in range(len(Y)):
            peak_height = aux.find_peak_height(X=X, Y=Y[pos], x_min=x_min, x_max=x_max)
            if pos in peak_heights_dict:
                peak_heights_dict[pos].append(peak_height)
            else:
                peak_heights_dict[pos] = [peak_height]

    # Calculate the average peak height for each position
    positions = list(peak_heights_dict.keys())
    avg_peak_heights = [np.mean(peak_heights_dict[pos]) for pos in positions]

    # Plot the average peak heights as a function of position
    fig, ax = plt.subplots()
    ax.minorticks_on()
    plt.plot(
        positions,
        avg_peak_heights,
        marker=".",
        linestyle="-",
        color="royalblue",
        alpha=0.9,
    )
    plt.xlabel("Position")
    plt.ylabel("Average Maximum Peak Height")
    plt.title(
        f"Exp: {dataset.fl_num}, height group: {dataset.height_group},peak range=[{x_min},{x_max}], scan_num = {scan_number}"
    )
    plt.show()

    # Export the data to an Excel file if requested
    if export_table:
        df_export = pd.DataFrame(
            {"Position": positions, "Average Peak Height": avg_peak_heights}
        )
        df_export.to_excel(export_table, index=False)



"""
This module contains functions that load and handles parsing of the raw data.

Cerntain information specific to experimental files obtaining from a beamtime during 14-18 th  September 2023 was
incoperated here for smooth parsing of the raw data. For other set of experiment, please adjust these conditions
acordingly.
"""
import pandas as pd
import h5py
import numpy as np
from pathlib import Path
from typing import Dict, Any, List, Tuple, Optional
import plotly.graph_objects as go
from IPython.display import display, clear_output
from ipywidgets import interactive, SelectionSlider, IntSlider, Checkbox
from . import auxiliary as aux


class LoadData:
    """
    This class takes an experiment number and height group to create an object that processes properties required for
    other functions used for analysis and visualizing data.

    This class also includes a show_spectrum class function, allowing users to quickly visualize the spectrum.
    """

    def __init__(self, fl_num: int, height_group: int, data_info: Optional[str] = None):
        self.fl_num = fl_num
        self._height_group = height_group
        self.data_info = data_info
        (
            self._fl_integrated,
            self._fl_raw,
            self._fl_start_macro,
            self._fl_end_macro,
            self._condition,
            self._pos_scan_motor,
            self._h_group_motor,
        ) = self.get_fl_detail()

    def get_fl_detail(
        self,
        input_fl_integrated: Optional[str] = None,
        input_fl_raw: Optional[str] = None,
        input_fl_start_macro: Optional[int] = None,
        input_fl_end_macro: Optional[int] = np.nan,
        input_condition: Optional[str] = None,
        input_pos_scan_motor: Optional[str] = "pp01",
        input_h_group_motor: Optional[str] = "h1tz",
    ) -> Tuple[str, str, int, Optional[int], str, str, str]:
        """
        Retrieves file details based on the experimental number.

        If data_info is provided, it fetches details from the Excel file, otherwise, it uses the provided input values.
        """
        if self.data_info:
            try:
                df_exp_info = pd.read_excel(self.data_info)
            except FileNotFoundError:
                raise Exception(f"Excel file not found: {self.data_info}")

            data_row = df_exp_info[df_exp_info["Experimental number"] == self.fl_num]

            if data_row.empty:
                raise ValueError(
                    f"Experimental number {self.fl_num} not found in the data."
                )

            fl_integrated_path = Path(
                data_row["Integrated file directory"].values[0]
            ).resolve()
            fl_raw_path = Path(data_row["Raw file directory"].values[0]).resolve()

            fl_start_macro = data_row["Macro start number"].values[0]
            fl_end_macro = (
                data_row["Macro end number (optional)"].values[0]
                if not pd.isna(data_row["Macro end number (optional)"].values[0])
                else None
            )
            condition = data_row["Condition name"].values[0]
            pos_scan_motor = data_row["Position scanning motor"].values[0]
            h_group_motor = data_row["Height group motor"].values[0]
        else:
            fl_integrated_path = (
                Path(input_fl_integrated).resolve() if input_fl_integrated else None
            )
            fl_raw_path = Path(input_fl_raw).resolve() if input_fl_raw else None
            fl_start_macro = input_fl_start_macro
            fl_end_macro = input_fl_end_macro
            condition = input_condition
            pos_scan_motor = input_pos_scan_motor
            h_group_motor = input_h_group_motor

            if not all(
                [
                    fl_integrated_path,
                    fl_raw_path,
                    fl_start_macro,
                    condition,
                    pos_scan_motor,
                    h_group_motor,
                ]
            ):
                raise ValueError(
                    "All file details must be provided if not using an Excel file."
                )

        return (
            str(fl_integrated_path),
            str(fl_raw_path),
            fl_start_macro,
            fl_end_macro,
            condition,
            pos_scan_motor,
            h_group_motor,
        )

    @property
    def fl_raw(self) -> str:
        return self._fl_raw

    @property
    def height_group(self) -> int:
        return self._height_group

    @property
    def fl_integrated(self) -> str:
        return self._fl_integrated

    @property
    def fl_start_macro(self) -> int:
        return self._fl_start_macro

    @property
    def fl_end_macro(self) -> Optional[int]:
        return self._fl_end_macro

    @property
    def pos_scan_motor(self) -> str:
        return self._pos_scan_motor

    @property
    def h_group_motor(self) -> str:
        return self._h_group_motor

    def get_max_frame_index(self) -> int:
        with h5py.File(self.fl_raw, "r") as f:
            keys = list(f.keys())
            frame_indices = [int(key.split(".")[0]) for key in keys if "." in key]
            return max(frame_indices)

    def get_height_array(self, start: int, num_end: int) -> Dict[int, Any]:
        if pd.isna(num_end):
            end = self.get_max_frame_index()
        else:
            end = int(num_end)
        dict_height = {}
        for scan_num in range(start, end + 1):
            height = aux.get_data(
                fl=self._fl_raw,
                dataset_path=f"{scan_num}.1/instrument/positioners/{self.h_group_motor}",
            )
            dict_height[scan_num] = height
        return dict_height

    @property
    def height_group_frame(self) -> Dict[int, Any]:
        height_array = self.get_height_array(self._fl_start_macro, self._fl_end_macro)
        grouped_height_array = group_heights(height_array)
        self._height_group_frame = grouped_height_array[self.height_group]
        return self._height_group_frame

    def show_spectrum(self, xref_list: Optional[List] = None):
        # Function to draw vertical bars and legend labels
        def ybar_plotly(fig, x, label, thick=0.02, alpha=0.25, color="green"):
            fig.add_shape(
                type="rect",
                x0=x - thick,
                x1=x + thick,
                y0=0,
                y1=1,
                fillcolor=color,
                opacity=alpha,
                line=dict(width=0),
                xref="x",
                yref="paper",
            )
            if label:  # Add a legend entry only if a label is provided
                fig.add_trace(
                    go.Scatter(
                        x=[None],
                        y=[None],
                        mode="markers",
                        marker=dict(color=color),
                        name=label,
                        showlegend=True,
                        legendgroup=label,
                    )
                )

        # Initialize your figure
        fig = go.FigureWidget()
        fig.update_layout(
            title={
                "text": f"Exp: {self.fl_num}, Height group: {self.height_group} <br> Condition: {self._condition}",
                "y": 0.9,
                "x": 0.5,
                "xanchor": "center",
                "yanchor": "top",
            },
            title_font=dict(size=20),
            xaxis_title="q",
            yaxis_title="Count",
            width=1600,
            height=600,
            legend=dict(
                x=1,  # Legend x position
                y=1,  # Legend y position
                traceorder="normal",
                font=dict(family="sans-serif", size=15, color="black"),
                bgcolor="White",
                bordercolor="White",
                borderwidth=0,
            ),
        )

        # Add a primary trace for the data
        fig.add_trace(
            go.Scatter(
                x=[],
                y=[],
                mode="lines",
                line=dict(color="red", width=2),
                name="Spectrum",
                opacity=0.7,
                showlegend=False,
            )
        )

        # If xref_list is provided, add bars and legends
        if xref_list is not None:
            for item in xref_list:
                xrefs, color, label = item
                added_legend = False
                for pos in xrefs:
                    ybar_plotly(
                        fig,
                        pos,
                        label if not added_legend else "",
                        thick=0.01,
                        color=color,
                    )
                    added_legend = True

        display(fig)

        # Function to update the plot with new data
        def plot_data(
            n: int, position: int, bg_substract: bool, log_scale: bool
        ) -> None:
            x = aux.get_data(
                fl=self._fl_integrated, dataset_path=f"{n}.1/p3_integrate/integrated/q"
            )
            y_0 = aux.get_data(
                fl=self._fl_integrated,
                dataset_path=f"{self.height_group_frame[0]}.1/p3_integrate/integrated/intensity",
            )[position]
            y_n = aux.get_data(
                fl=self._fl_integrated,
                dataset_path=f"{n}.1/p3_integrate/integrated/intensity",
            )[position]

            y = y_n - y_0 if bg_substract else y_n

            # Update the data in the existing figure
            fig.data[0].x = x
            fig.data[0].y = y
            fig.update_yaxes(type="log" if log_scale else "linear")
            clear_output(wait=True)

        max_positions = 0
        for n in self.height_group_frame:
            Y = aux.get_data(
                fl=self._fl_integrated,
                dataset_path=f"{n}.1/p3_integrate/integrated/intensity",
            )
            max_positions = max(max_positions, len(Y))

        # Create interactive sliders and checkbox
        interactive_plot = interactive(
            plot_data,
            n=SelectionSlider(
                options=self.height_group_frame, description="Scan number"
            ),
            position=IntSlider(
                min=0, max=max_positions - 1, step=1, description="Position"
            ),
            bg_substract=Checkbox(value=False, description="Background Subtraction"),
            log_scale=Checkbox(value=False, description="Log Scale Y"),
        )
        display(interactive_plot)


def group_heights(h_array: Dict[int, int]) -> Dict[int, int]:
    """This function sort the peak the number of peak position at to 4 groups (3 experimental height and 1 controlled
    height)"""
    rounded_h_array = {k: round(v, 3) for k, v in h_array.items()}
    groups = {}
    for scan_num, height in rounded_h_array.items():
        if height not in groups:
            groups[height] = []
        groups[height].append(scan_num)
    unique_heights = sorted(groups.keys())
    h_list = [groups[height] for height in unique_heights]
    control_height = min(unique_heights, key=lambda x: len(groups[x]))
    for i, height in enumerate(unique_heights):
        height_type = "control height" if height == control_height else f"height {i}"
        # print(f"list position {i}; {height_type} at {height} mm")
    return h_list



from setuptools import setup, find_packages

setup(
    name="twaxs",
    version="2.1.0",
    package_dir={'': 'src'},
    author="Nukorn Plainpan",
    author_email="nukorn.p27@gmail.com",
    packages=find_packages(where='src'),
    install_requires=[
    'h5py',
    'matplotlib',
    'ipywidgets',
    'IPython',
    'plotly',
    'numpy',
    'pandas',
    'scipy',
    'python-dateutil',
    'openpyxl',
    'ipykernel',
    'pathlib'
    ],
)



"""This module contains functions that helps analyze the raw data which can be used by the plotting functions in the
plot module."""

import h5py
import numpy as np
import pandas as pd
import os
import math
from dateutil.parser import parse
from scipy.signal import savgol_filter
from typing import List, Union, Optional, Any


def get_data(fl: str, dataset_path: str) -> Any:
    """
    Extract data from a specified hdf5 file and dataset path.

    Parameters:
    fl (str): The file path for the hdf5 file.
    dataset_path (str): The specific dataset path within the hdf5 file.

    Returns:
    Any: The data found at the specified dataset path within the file.
    """
    with h5py.File(fl, "r") as f:
        data = f[dataset_path][()]
    return data


def find_peak_height(
    X: Union[List[float], np.ndarray],
    Y: Union[List[float], np.ndarray],
    x_min: float,
    x_max: float,
) -> Optional[float]:
    """
    Find the peak height in a given XRD diffractogram within a specified x range.

    Parameters:
    X (list or numpy array): The array of x values (2 theta or similar)
    Y (list or numpy array): The array of y values (intensity)
    x_min (float): The minimum x value of the range to consider
    x_max (float): The maximum x value of the range to consider

    Returns:
    float: The peak height in the given range, or None if the range is invalid
    """
    # Converting X and Y to numpy arrays if they are not already
    X = np.array(X)
    Y = np.array(Y)

    # Finding the indices of X values within the range [x_min, x_max]
    valid_indices = np.where((X >= x_min) & (X <= x_max))

    if len(valid_indices[0]) == 0:
        return None

    # Finding the peak height within the specified range
    peak_height = np.max(Y[valid_indices])

    return peak_height


def get_peak_height_time(
    dataset: "LoadData",
    x_min: float,
    x_max: float,
    position: int,
    smoothing_window: Union[None, int] = None,
    n_pol: int = 2,
) -> pd.DataFrame:
    """This function returns a maximum peak height in the speciifc area at the given position as a function of time
    stamp."""
    df_hight_time = pd.DataFrame(columns=["time", "peak height"])
    raw_data = dataset.fl_raw
    frame = dataset.height_group_frame
    integrated_data = dataset.fl_integrated
    for n in frame:
        time = get_scan_time(raw_data, scan_num=n)
        intensity_data = get_data(
            fl=integrated_data, dataset_path=f"{n}.1/p3_integrate/integrated/intensity"
        )[position]

        # If smoothing is desired, apply the Savitzky-Golay filter
        if smoothing_window:
            # Ensure the window size is odd
            if smoothing_window % 2 == 0:
                smoothing_window += 1
            intensity_data = savgol_filter(intensity_data, smoothing_window, n_pol)

        peak_height = find_peak_height(
            X=get_data(
                fl=integrated_data, dataset_path=f"{n}.1/p3_integrate/integrated/q"
            ),
            Y=intensity_data,
            x_min=x_min,
            x_max=x_max,
        )
        df_hight_time.loc[len(df_hight_time)] = [time, peak_height]

    return df_hight_time


def get_scan_time(fl: str, scan_num: int) -> float:
    """Get the time stamp of the scan in utx."""
    exp_timestamp = get_data(fl=fl, dataset_path=f"{scan_num}.1/start_time").decode(
        "utf-8"
    )
    exp_timestamp_epoc = float(parse(exp_timestamp).timestamp())
    return exp_timestamp_epoc


def get_fe(path_gc_excel: str) -> pd.DataFrame:
    """This function take the excel fe path (path)gc_excel) and return an array of a dataframe containing utx time stamp
    and the Faradaic efficiency of different gas product."""
    input_df = pd.read_excel(path_gc_excel)

    # Helper function get_col
    def get_col(input_df: pd.DataFrame, start_row: int = 2) -> List:
        """
        Extracts a column from input_df starting from start_row.

        :param input_df: The dataframe to process.
        :param start_row: The row index to start the extraction from.
        :return: A list representing the extracted column.
        """
        trans_col = []
        for idx in range(start_row, len(input_df)):
            trans_col.append(input_df[idx])
        return trans_col

    for idx_col, col_name in enumerate(input_df.columns):
        if col_name == "fe":
            idx_fe = idx_col
    for idx_col_findFeEnding in range(idx_fe + 1, len(input_df.columns)):
        col_name = input_df.columns[idx_col_findFeEnding]
        if col_name.split(":")[0] != "Unnamed":
            fe_endRange = idx_col_findFeEnding
            break
    df_feGC = pd.DataFrame()
    for idx_col_findChemicalName in range(idx_fe, fe_endRange):
        col_name = input_df.iloc[0, idx_col_findChemicalName]
        col_data = get_col(input_df.iloc[:, idx_col_findChemicalName])
        df_feGC.insert(len(df_feGC.columns), col_name, col_data)
    df_feGC.insert(len(df_feGC.columns), "Overall", df_feGC.sum(axis=1))
    df_feGC.insert(0, "time", get_col(input_df["Unnamed: 0"]))

    return df_feGC


def export_spectrum(
    data: "LoadData", scan_num: int, position: int, export_dir: Union[bool, str] = False
) -> pd.DataFrame:
    """This function print the q and count value of the spectrum of a specific scan number and position."""
    from . import dataset

    q = get_data(
        fl=data.fl_integrated, dataset_path=f"{scan_num}.1/p3_integrate/integrated/q"
    )
    count = get_data(
        fl=data.fl_integrated,
        dataset_path=f"{scan_num}.1/p3_integrate/integrated/intensity",
    )[position]
    df_export = pd.DataFrame({"q": q, "count": count})
    if export_dir is not False:
        df_export.to_excel(export_dir, index=False)
    return df_export


def twotheta2q(angle: float, wavelength: float = 1.5406) -> float:
    """
    This function take the 2theta as in input and convert them to a respective q vector.

    Args:
        angle (float): 2 Theta angle from an xray diffraction pattern.
        wavelength (float, optional): wavelength of the xray source of the input spectrum. The default is set to 1.5406
        which is the xray wavelength of an Cu k alpha.

    Returns:
        float: converted q vector value.
    """
    theta = math.radians(angle / 2)
    q = (4 * math.pi * math.sin(theta)) / wavelength
    return q


def get_ref_xray(excel_xray_fl: str, threshold: int = 10) -> List[float]:
    """Exract the diffraction pattern with intensity higher than the set treshold from a standard diffraction
    pattern."""
    df_xray = pd.read_excel(excel_xray_fl)
    xray_array = []
    for idx in range(0, len(df_xray)):
        if df_xray["I [%]"][idx] >= threshold:
            xray_array.append(twotheta2q(df_xray["2θ [°]"][idx]))
    return xray_array


def load_xray_ref_folder(
    xray_ref_folder: str, threshold: int = 10
) -> List[List[float]]:
    """
    Load X-ray reference files from a specified folder and convert them into a list of X-ray patterns.

    This function reads all files in the given folder path, converts the X-ray reference data in each file
    into a list of X-ray patterns using the get_ref_xray function, and aggregates them into a list.

    Args:
    xray_ref_folder (str): Path to the folder containing the X-ray reference files.
    threshold (int, optional): A cutoff intensity percentage; patterns with intensity below this threshold
    will be excluded. Defaults to 10.

    Returns:
    List[List[float]]: A list containing lists of X-ray patterns from each file.
    """
    xray_ref_list = []
    for file in sorted(os.listdir(xray_ref_folder)):
        if file.endswith(".xlsx"):  # assuming the files are .xlsx, adjust if different
            full_file_path = os.path.join(xray_ref_folder, file)
            xray_ref_list.append(get_ref_xray(full_file_path, threshold=threshold))
    return xray_ref_list


def get_height_diff(dataset: "LoadData") -> float:
    """
    Get the average step different in position (in micrometer).

    Parameters:
    dataset: dataset object of the associated experiment.

    Returns:
    float: Average height difference.
    """
    arr_position = get_data(
        fl=dataset.fl_raw,
        dataset_path=f"{dataset.fl_start_macro}.1/instrument/positioners/{dataset.pos_scan_motor}",
    )
    pos_last = arr_position[-1]
    post_first = arr_position[0]
    diff = abs(pos_last - post_first) * 1000 / len(arr_position)
    return diff
